
// This file is auto-generated during compilation, do not modify
#[derive(RustEmbed)]
#[folder = r"/build/smolvm/src/smolvm/target/release/build/utoipa-swagger-ui-adbc866158517af3/out/swagger-ui-5.17.14/dist/"]
struct SwaggerUiDist;
